/*
 * screenload - Loads SUN rasterfiles onto the ABC1600 console screen.
 *              Handles standard and run length encoded files with pixel
 *              depths up to eight bits (color map images).
 *
 *              NB: The program has to be setuid root!
 *
 *
 * Author: jonas-y (Jonas Yngvesson, jonas-y@isy.liu.se) 1991/92
 *
 * This program is Public Domain.
 *
 *
 * usage: screenload [filename]
 *
 * If no filename is provided data is read from standard input.
 */
 
#include <stdio.h>
#include <dnix/fcodes.h>
#include "rasterfile.h"         /* File pinched from SUN */

#define ERASE_SCREEN() puts("\33:0;0m\33:767;1023;0;0f")


/*
 * abcenix cc doesn't grok void, sigh...
 */
typedef int void;


/*
 * Structure used when mapping frame buffer into user memory.
 */
struct {
    unsigned char *addr;
    int            size;
} frame_buf;


/*
 * Other global variables.
 */
struct rasterfile rf;             /* Rasterfile header */
unsigned char    *colormap[3];    /* Colormap */
int               rle;            /* True if image is run length encoded */
int               ncolors;        /* Number of entrys in colormap */
int               greyscale;      /* True if its a greyscale image */

unsigned char     redtab[256];    /* lookup tables for greyscale conversion */
unsigned char     grntab[256];
unsigned char     blutab[256];

char 		  alphatab[256];  /* lookup tables for halftoning */
char		  betatab[256];
char     	  gammatab[256];
char     	  deltatab[256];


/*
 * Function for reading one scanline of image data.
 * Also handles run length decoding.
 *
 * Run length encoded chunks are not aligned on scanlines
 * so the function must save its state in static variables
 * between calls to be able to continue decoding correctly.
 */
void
read_line(infile, buf, len)
    FILE *infile;
    unsigned char *buf;
    int len;
{
    static int left = 0; /* Number of bytes left in current rle chunk */
    static int code;     /* Current rle byte */
    int        bufpos;

    /*
     * If data isn't run length encoded we just read it in.
     */
    if (!rle) {
        fread(buf, 1, len, infile);
	return;
    }
    
    /*
     * Decode LEN bytes.
     */
    bufpos = 0;
    while (bufpos != len) {
        if (left) {
	    buf[bufpos++] = code;
	    left--;
        } else {
	    code = getc(infile);
	    if ((code == 0x80) && ((left = getc(infile)) != 0)) {
  	        code = getc(infile);
	        left++;
	    } else {
	        left = 1;
	    }
	}
    }
}


/*
 * Display a bitmap image.
 */
void
bitmap_display(infile)
    FILE *infile;
{
    unsigned char *bufpos;      /* Position in frame buffer of scanline start */
    int            line_length; /* Length of a line in bytes (in the file) */
    int            width;       /* Width of the visible part of the image */
    int            scrapped;    /* line_length - width */
    unsigned char *scrap_buf;   /* Buffer for scrapped bytes */
    int            tmp;         /* Used in calculation of bufpos */

    /*
     * Since lines are padded in the file to be
     * multiple of 16 bits, line_length is not
     * certain to be equal to rf.ras_width / 8.
     */
    line_length = (rf.ras_width + 7) >> 3;
    line_length += (line_length & 1);

    /*
     * Center image on the screen or, if it's to large,
     * decide which parts of it that is visible.
     */
    tmp = (1024 - rf.ras_height) / 2;   /* Center of screen in y direction */
    tmp = (tmp < 0 ? 0 : tmp);
    bufpos = frame_buf.addr + tmp * 128;
    
    tmp = (768 - rf.ras_width) / 2;     /* Center of screen in x direction */
    if (tmp < 0) {
	width = 96;
	scrapped = line_length - 96;
        scrap_buf = (unsigned char *)malloc(scrapped * sizeof(unsigned char));
    } else {
        bufpos += tmp >> 3;
        width = line_length;
	scrapped = 0;
    }
    
    /*
     * Read one line at the time and write into the framebuffer
     */
    while (rf.ras_height && bufpos < frame_buf.addr + 128 * 1024) {
        read_line(infile, bufpos, width);
        if (scrapped) {
	    read_line(infile, scrap_buf, scrapped);
	}
	bufpos += 128;
	rf.ras_height--;
    }
}


/*
 * Read the colormap (if any).
 * Init lookuptables for color to greyscale conversion.
 *
 * If r, g and b are equal in all colormap entrys we
 * assume the image is greyscale already.
 */
void
color_convert_init(infile)
    FILE *infile;
{
    int i;
    
    if (rf.ras_maplength != 0) {
        ncolors = rf.ras_maplength / 3;
	for (i = 0; i < 3; i++) {
	    colormap[i] = (unsigned char *)malloc(ncolors);
	    fread(colormap[i], 1, ncolors, infile);
	}
    }
    
    greyscale = 0;
    for (i = 0; i < ncolors; i++) {
        greyscale = (colormap[0][i] == colormap[1][i]
                     && colormap[1][i] == colormap[2][i]);
	if (!greyscale) {
	    break;
	}
    }

    /*
     * NTSC formula for color to grayscale conversion.
     */
    if (!greyscale) {
        for (i = 0; i < 256; i++) {
            redtab[i] = 0.30 * i;
	    grntab[i] = 0.59 * i;
	    blutab[i] = 0.11 * i;
	}
    }
}    
    
    
/*
 * Init lookuptables for error distribution (Floyd-Steinberg algorithm)
 */
halftone_init()
{
    int i;

    /* Distribute the error as follows: */
    /*    current pixel = A[y,x]        */
    /*    A[y,   x+1] += err*7/16       */
    /*    A[y+1, x-1] += err*3/16       */
    /*    A[y+1, x  ] += err*5/16       */
    /*    A[y+1, x+1] += err*1/16       */
    for (i = 0; i < 256; i++) {
	alphatab[i] = ((i - 128) * 7) / 16;
	betatab[i]  = ((i - 128) * 3) / 16;
	gammatab[i] = ((i - 128) * 5) / 16;
	deltatab[i] = ((i - 128) * 1) / 16;
    }
}


/*
 * Display a color or greyscale image.
 */
void
halftone_display(infile)
    FILE *infile;
{
    register unsigned int code;  /* Bit pattern after halftoning. */
    register unsigned int mask;  /* Bit mask used when setting bits in CODE */
    register short  grey;        /* Grey level in current pixel */
    unsigned int   *bufpos;      /* Position in frame buffer of scanline start */
    int             line_length; /* Length of a line in bytes (in the file) */
    int             offset;      /* Offset to pixel from BUFPOS */
    unsigned char  *databuf;     /* Image data read from file */
    register unsigned char *dp;  /* Offset into DATABUF */
    char           *errbuf1;     /* Buffer used for error distribution */
    char           *errbuf2;     /* d.o. */
    register char  *errp1;       /* Pointer into error buffer */
    register char  *errp2;       /* Pointer into error buffer */
    int             width;       /* size of visible part of lines */
    char           *tmp2;
    int             tmp;
    register int    i;
    

    /*
     * Since lines are padded in the file to be
     * multiple of 16 bits, line_length is not
     * certain to be equal to rf.ras_width.
     */
    line_length = rf.ras_width + (rf.ras_width & 1);
    
    /* 
     * Center image on screen.
     */
    tmp = (1024 - rf.ras_height) / 2;   /* Center of screen in y direction */
    tmp = (tmp < 0 ? 0 : tmp);
    (unsigned char *)bufpos = frame_buf.addr + tmp * 128;
    
    tmp = (768 - rf.ras_width) / 2;     /* Center of screen in x direction */
    if (tmp < 0) {
	width = 768;
    } else {
        bufpos += (tmp >> 5);      /* Align on int */
        width = line_length;
    }

    /*
     * Allocate buffers for the image data and
     * error distribution.
     */
    databuf = (unsigned char *)malloc(line_length);
    errbuf1 = (char *)calloc(sizeof(char), width + 2);
    errbuf2 = (char *)calloc(sizeof(char), width + 2);
    errp1 = errbuf1 + 1;
    errp2 = errbuf2;


    /*
     * Read one line at the time, do halftoning of it
     * and write into the framebuffer
     */
    while (rf.ras_height 
           && (unsigned char *)bufpos < frame_buf.addr + 128 * 1024) {
        read_line(infile, databuf, line_length);
        dp = databuf;
	code = 0;
	mask = 0x80000000;
        offset = 0;
	for (i = width; i; i--) {
	    grey = *dp++;
	    if (!greyscale) {
	        grey = (redtab[colormap[0][grey]] 
		        + grntab[colormap[1][grey]] 
			+ blutab[colormap[2][grey]]);
	    } else {
	        grey = colormap[0][grey];
	    }

	    if ((grey += *errp1) < 128) {
		code |= mask;
		grey += (short)128;
	    } else {
	        grey -= (short)128;
	    }
	    
	    *errp1++  = 0;
	    *errp1   += alphatab[grey];
	    *errp2++ += betatab[grey];
	    *errp2++ += gammatab[grey];
	    *errp2-- += deltatab[grey];
	    
 	    if ((mask >>= 1) == 0) {
	        *(bufpos + offset++) = code;
		code = 0;
	        mask = 0x80000000;
	    }
	}
	
	if (mask != 0x80000000) {
	    *(bufpos + offset) = code;
	}

	tmp2 = errbuf1;
	errbuf1 = errbuf2;
	errbuf2 = tmp2;
	errp1 = errbuf1 + 1;
	errp2 = errbuf2;

        bufpos += 32;
	rf.ras_height--;
    }
}



main(argc, argv)
int argc;
char *argv[];
{
    FILE *infile;

    if (argc == 1) {
        infile = stdin;
    } else if ((infile = fopen(argv[1], "r")) == NULL) {
        perror("Couldn't open rasterfile");
	exit(1);
    }

    /*
     * Read the rasterfile header and check if
     * it is correct.
     */
    fread(&rf, 1, sizeof(rf), infile);
    if (rf.ras_magic != RAS_MAGIC) {
        fputs("Not a valid rasterfile.\n", stderr);
	exit(1);
    }
    
    /*
     * Check the type of image.
     */
    if (rf.ras_type != RT_OLD 
        && rf.ras_type != RT_STANDARD
        && rf.ras_type != RT_BYTE_ENCODED)
    {
        fprintf(stderr,"Unsupported rasterfile type: %d.\n", rf.ras_type);
	exit(1);
    }
    rle = (rf.ras_type == RT_BYTE_ENCODED);
    
    /*
     * Map in the frame buffer.
     */
    if (dnix(F_SEGP, &frame_buf, sizeof(frame_buf), 3, 0) != sizeof(frame_buf)) {
        fputs("Couldn't map in frame buffer. Are you root?\n", stderr);
        exit(1);
    }

    ERASE_SCREEN();

    /*
     * If it is a color file, we display with
     * Floyd-Steinberg halftoning.
     * Otherwize its a plain bitmap ready for display.
     */
    if (rf.ras_depth != 1) {
        color_convert_init(infile);
	halftone_init();
	halftone_display(infile);

    } else {
        bitmap_display(infile);
    }
}

